theory COComp
  imports TransitionSystems UnifiedAxioms
begin

context RWVarsActions
begin

type_synonym T = nat
type_synonym R = nat
type_synonym L = nat
type_synonym V = nat
type_synonym RF = "R \<Rightarrow> V"


datatype 'pc AComm = (* Control location *)
      loadc bool R L 'pc
    | storec bool L V 'pc
    | storee bool L "RF \<Rightarrow> V" 'pc
    | test "RF set" 'pc 'pc 
    | skip
    | fencec R V 'pc 
    | loada bool R L R V 'pc (* Auxiliary state (register location aux_register value) *)


datatype mem_op =
   load bool L R V | store bool L V | fence_op R V | load_aux bool L R V R V



fun comm_step :: "(T \<Rightarrow> 'pc \<Rightarrow> 'pc AComm) \<Rightarrow> T \<times> (mem_op option) \<Rightarrow> (RF \<times> 'pc) rel"
  where
  "comm_step Pr (t, None) =
               {((rf, pc), (rf, if (rf \<in> B) then pc_t else pc_f)) | rf pc B pc_t pc_f.
                                    Pr t pc = test B pc_t pc_f}" |
  "comm_step Pr (t, Some (load b x r v)) =
               {((rf, pc), (rf(r := v), pc')) | rf pc pc'.
                                    Pr t pc = loadc b r x pc'}" |
  "comm_step Pr (t, Some (store b x v)) =
               {((rf, pc), (rf, pc')) | rf pc pc'.
                                    Pr t pc = storec b x v pc'} \<union>
               {((rf, pc), (rf, pc')) | e rf pc pc'. v = e rf \<and>
                                    Pr t pc = storee b x e pc'}" | 
  "comm_step Pr (t, Some (fence_op r v)) =
               {((rf, pc), (rf(r := v), pc')) | rf pc pc'.
                                    Pr t pc = fencec r v pc'}" |
  "comm_step Pr (t, Some (load_aux b x r v ra va)) =
               {((rf, pc), (rf(r := v, ra := va), pc')) | rf pc pc'.
                                    Pr t pc = loada b r x ra va pc'}"

lemma test_step_true:
  assumes "Pr t pc = test B pc_t pc_f"
      and "rf \<in> B"
    shows "((rf, pc), (rf, pc_t)) \<in> comm_step Pr (t, None)"
  using assms by force

lemma test_step_false:
  assumes "Pr t pc = test B pc_t pc_f"
      and "rf \<notin> B"
    shows "((rf, pc), (rf, pc_f)) \<in> comm_step Pr (t, None)"
  using assms by force

lemma test_step_elim:
  assumes "Pr t pc = test B pc_t pc_f"
      and "((rf, pc), (rf', pc')) \<in> comm_step Pr (t, None)"
    shows "rf' = rf \<and> ((rf \<in> B \<and> pc' = pc_t) \<or> (rf \<notin> B \<and> pc' = pc_f))"
  using assms by simp


definition
  "par_step Ts A tao \<equiv>
      let (t, ao) = tao in {(ls, ls(t := l')) | ls l'. t \<in> Ts \<and> (ls t, l') \<in> A tao}"

definition
  "co_step m S C e \<equiv>
      let (t, ao) = e in
      case ao of
        Some a \<Rightarrow> {((s, c), (s', c')) | c s c' s' o_act.
          m a o_act \<and> (s, s') \<in> S (t, o_act) \<and> (c, c') \<in> C (t, Some a)}
      | None \<Rightarrow> {((s, c), (s, c')) | c s c'. (c, c') \<in> C (t, None)}"


end


context RWVarsActions
begin

definition mem_act :: "mem_op \<Rightarrow> 'a \<Rightarrow> bool"
  where
  "mem_act op a \<equiv>
     case op of
        (load b x r v) \<Rightarrow> a = R b x v |
        (store b x v) \<Rightarrow> a = W b x v |
        (fence_op r v) \<Rightarrow> a = F |
        (load_aux b x r v ra va) \<Rightarrow> a = R b x v" 
(* |
        (mem_set br bw x v') \<Rightarrow> \<exists> v. a = U br bw x v v'" *)

definition "full_step Ts Pr = co_step mem_act G (par_step Ts (comm_step Pr))"

lemma full_step_update_other:
  assumes "(s, s') \<in> full_step Ts Pr e"
      and "fst e \<noteq> t"
    shows "snd s' t = snd s t"
  apply(cases "snd e")
  using assms by(auto simp add: full_step_def par_step_def co_step_def)

type_synonym 'pc prog_type = "nat \<times> (mem_op option) \<Rightarrow> (((nat \<Rightarrow> nat) \<times> 'pc) \<times> 'pc) set"

type_synonym ('s, 'pc) full_state = "('s \<times> (nat \<Rightarrow> ((nat \<Rightarrow> nat) \<times> 'pc)))"

definition "lift_t t P = {ls | ls. fst(ls t) \<in> P}"

definition "lift\<^sub>1 P = {(s, c) | s c. s \<in> P}"
definition "lift\<^sub>2 P = {(s, c) | s c. c \<in> P}"

definition "thr_pc s t = snd (snd s t)"
definition "thr_rf s t = fst (snd s t)"

definition "inv_of Ts A = {s . (\<forall> t. t \<in> Ts \<longrightarrow> s \<in> A t (thr_pc s t))}"

definition
  "local_correctness Ts Pr A t \<equiv>
    \<forall> s ao s'. s \<in> A t (thr_pc s t) \<and> (s, s') \<in> full_step Ts Pr (t, ao) \<longrightarrow> s' \<in> A t (thr_pc s' t)"


definition "Reg t B = lift\<^sub>2(lift_t t B)"

abbreviation "reg r v \<equiv> {reg . reg r = v}"

lemma full_step_tau_step_ext:
  assumes "t \<in> Ts"
  shows "((s, ls), (s', ls')) \<in> full_step Ts Pr (t, None) \<longleftrightarrow>
             s' = s \<and> (\<exists> B pc_t pc_f. Pr t (snd(ls t)) = test B pc_t pc_f \<and>
                          ((fst(ls t) \<in> B \<and> ls' = ls(t := (fst(ls t), pc_t))) \<or> 
                           (fst(ls t) \<notin> B \<and> ls' = ls(t := (fst(ls t), pc_f)))))"
  using assms apply(auto simp add: full_step_def co_step_def par_step_def mem_act_def thr_pc_def)
  by force+

definition "arb_pc t = {((s, ls), (s, ls(t := (fst(ls t), pc')))) | s ls pc'. True}"

lemma arb_pc_Id:
  "Id \<subseteq> arb_pc t"
  apply(simp add: arb_pc_def) apply clarsimp
  by (metis fun_upd_triv prod.collapse)

lemma arb_pc_trans:
  "arb_pc t O arb_pc t \<subseteq> arb_pc t"
  apply(simp add: arb_pc_def) apply clarsimp
  by blast

lemma full_step_subs_arb_pc:
  "full_step Ts Pr (t, None) \<subseteq> arb_pc t"
  by(auto simp add: full_step_def co_step_def par_step_def mem_act_def arb_pc_def)

lemma arb_pc_full_step:
  assumes "Q \<subseteq> \<langle>arb_pc t\<rangle> Q"
      and "P \<subseteq> Q"
   shows "P \<subseteq> \<langle>full_step Ts Pr (t, None)\<rangle> Q"
  by (metis (no_types, lifting) assms(1) assms(2) dual_order.trans full_step_subs_arb_pc wp_act_antimono2)

lemma Reg_pres_arb_pc:
  shows "Reg t B \<subseteq> \<langle>arb_pc t'\<rangle> (Reg t B)"
  by(auto simp add: Reg_def lift\<^sub>2_def lift_t_def arb_pc_def wp_ext)

lemma lift\<^sub>1_pres_arb_pc:
  shows "lift\<^sub>1 P \<subseteq> \<langle>arb_pc t\<rangle> (lift\<^sub>1 P)"
  by(auto simp add: Reg_def lift\<^sub>2_def lift\<^sub>1_def lift_t_def arb_pc_def wp_ext)


lemma store_step_disj:
  assumes "(s, s') \<in> full_step Ts Pr (t, Some (store b x v))"
  shows "Pr t (thr_pc s t) = storec b x v (thr_pc s' t) \<or>
          (\<exists> e. Pr t (thr_pc s t) = storee b x e (thr_pc s' t) \<and> e (thr_rf s t) = v)"
  using assms
  by(auto simp add: thr_rf_def lift\<^sub>2_def lift_t_def Reg_def mem_act_def full_step_def co_step_def par_step_def thr_pc_def)


lemma local_correctness_cases:
  assumes "\<And> t t' pc. t \<in> Ts \<Longrightarrow> t' \<in> Ts \<Longrightarrow> A t pc \<subseteq> \<langle>arb_pc t'\<rangle> (A t pc)"
      and "\<And> pc B pc_t pc_f.
                Pr t pc = test B pc_t pc_f \<Longrightarrow>
                A t pc \<inter> Reg t B \<subseteq> A t pc_t"
      and "\<And> pc B pc_t pc_f.
                Pr t pc = test B pc_t pc_f \<Longrightarrow>
                A t pc \<inter> Reg t (-B) \<subseteq> A t pc_f"
      and "\<And> pc b x r v pc'.
                Pr t pc = loadc b r x pc' \<Longrightarrow>
                A t pc \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> (A t pc')"
      and "\<And> pc b x v pc'.
                Pr t pc = storec b x v pc' \<Longrightarrow>
                A t pc \<subseteq> \<langle>full_step Ts Pr (t, Some (store b x v))\<rangle> (A t pc')"
      and "\<And> pc b r v pc'.
                Pr t pc = fencec r v pc' \<Longrightarrow>
                A t pc \<subseteq> \<langle>full_step Ts Pr (t, Some (fence_op r v))\<rangle> (A t pc')"
      and "\<And> pc b e x v pc'.
                Pr t pc = storee b x e pc' \<Longrightarrow>
                A t pc \<inter> Reg t {rf . e rf = v} \<noteq> {} \<Longrightarrow>
                A t pc \<inter> Reg t {rf . e rf = v} \<subseteq> \<langle>full_step Ts Pr (t, Some (store b x v))\<rangle> (A t pc')"
      and "\<And> pc b x r v ra va pc'.
                Pr t pc = loada b r x ra va pc' \<Longrightarrow>
                A t pc \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v ra va))\<rangle> (A t pc')"
    shows "local_correctness Ts Pr A t"
  apply(simp add: local_correctness_def) apply clarsimp
  apply(case_tac ao)
   defer
  apply(case_tac "ab")
     apply(auto simp add: full_step_def co_step_def par_step_def thr_pc_def mem_act_def)[1]
  using assms(4) apply(auto simp add: mem_act_def wp_ext full_step_def co_step_def par_step_def thr_pc_def)[1]
  apply blast
     apply(auto simp add: full_step_def co_step_def par_step_def thr_pc_def mem_act_def)[1]
  using assms(5) apply(auto simp add: mem_act_def wp_ext full_step_def co_step_def par_step_def thr_pc_def)[1]
    apply blast
  using assms(7) apply(auto simp add: lift\<^sub>2_def lift_t_def Reg_def mem_act_def wp_ext full_step_def co_step_def par_step_def thr_pc_def)[1]
   apply force
  using assms(6) apply(auto simp add: mem_act_def wp_ext full_step_def co_step_def par_step_def thr_pc_def)[1]
   apply blast
  using assms(8)
   apply(auto simp add: mem_act_def wp_ext full_step_def co_step_def par_step_def thr_pc_def)[1]
  apply blast

  apply(auto simp add: full_step_def co_step_def par_step_def thr_pc_def)[1]
  apply(subgoal_tac "A t pc \<subseteq> \<langle>(Reg t B)\<^sup>d O full_step Ts Pr (t, None)\<rangle> (A t pc_t)")
    apply(auto simp add: wp_ext diag_def Reg_def lift_t_def lift\<^sub>2_def lift\<^sub>1_def)[1]
  using full_step_tau_step_ext[where Ts=Ts and t=t and Pr=Pr]
  apply fastforce
  apply(rule converse_int_p_wp_diag_act) 
  apply(rule arb_pc_full_step)
  using assms(1) apply blast
   apply (simp add: assms(2))

  apply(subgoal_tac "A t pc \<subseteq> \<langle>(Reg t (-B))\<^sup>d O full_step Ts Pr (t, None)\<rangle> (A t pc_f)")
   apply(auto simp add: wp_ext diag_def Reg_def lift_t_def lift\<^sub>2_def lift\<^sub>1_def)[1]
  using full_step_tau_step_ext[where Ts=Ts and t=t and Pr=Pr]
  apply fastforce
   apply(rule converse_int_p_wp_diag_act) 
  apply(rule arb_pc_full_step)
  using assms(1) apply blast
  by (simp add: assms(3))

definition
  "noninterference Ts pr_step A t \<equiv>
    \<forall> t' s ao s'. t \<in> Ts \<and> t' \<noteq> t \<and> s \<in> A t (thr_pc s t) \<and> s \<in> A t' (thr_pc s t') \<and>
            (s, s') \<in> full_step Ts pr_step (t', ao) \<longrightarrow> 
                     s' \<in> A t (thr_pc s t)"

lemma noninterference_cases:
  assumes "\<And> t t' pc. t \<in> Ts \<Longrightarrow> t' \<in> Ts \<Longrightarrow> A t pc \<subseteq> \<langle>arb_pc t'\<rangle> (A t pc)"
      and "\<And> t' pc b x r v pc' pc''.
                t' \<in> Ts \<Longrightarrow>
                t' \<noteq> t \<Longrightarrow>
                Pr t' pc' = loadc b r x pc'' \<Longrightarrow>
                A t pc \<inter> A t' pc' \<subseteq> \<langle>full_step Ts Pr (t', Some (load b x r v))\<rangle> (A t pc)"
      and "\<And> t' pc b x v pc' pc''.
                t' \<in> Ts \<Longrightarrow>
                t' \<noteq> t \<Longrightarrow>
                Pr t' pc' = storec b x v pc'' \<Longrightarrow>
                A t pc \<inter> A t' pc' \<subseteq> \<langle>full_step Ts Pr (t', Some (store b x v))\<rangle> (A t pc)"
      and "\<And> t' pc b r v pc' pc''.
                t' \<in> Ts \<Longrightarrow>
                t' \<noteq> t \<Longrightarrow>
                Pr t' pc' = fencec r v pc'' \<Longrightarrow>
                A t pc \<inter> A t' pc' \<subseteq> \<langle>full_step Ts Pr (t', Some (fence_op r v))\<rangle> (A t pc)"
      and "\<And> t' pc b e x v pc' pc''.
                t' \<in> Ts \<Longrightarrow>
                t' \<noteq> t \<Longrightarrow>
                Pr t' pc' = storee b x e pc'' \<Longrightarrow>
                A t pc \<inter> A t' pc' \<inter> Reg t' {rf . e rf = v} \<subseteq> \<langle>full_step Ts Pr (t', Some (store b x v))\<rangle> (A t pc)"
      and "\<And> t' pc b x r v ra va pc' pc''.
                t' \<in> Ts \<Longrightarrow>
                t' \<noteq> t \<Longrightarrow>
                Pr t' pc' = loada b r x ra va pc'' \<Longrightarrow>
                A t pc \<inter> A t' pc' \<subseteq> \<langle>full_step Ts Pr (t', Some (load_aux b x r v ra va))\<rangle> (A t pc)"
    shows "noninterference Ts Pr A t"
  apply(case_tac "t \<notin> Ts")
  apply(fastforce simp add: noninterference_def)
  apply(simp add: noninterference_def) apply clarsimp
  apply(subgoal_tac "t' \<in> Ts")
  defer apply(case_tac ao; fastforce simp add: mem_act_def full_step_def co_step_def par_step_def thr_pc_def)[1]
  apply(case_tac ao)
   defer
   apply(case_tac "ab")
    apply(subgoal_tac "Pr t' (thr_pc (a, b) t') = loadc x11 x13 x12 (thr_pc (aa, ba) t')")
  using assms(2) apply (meson Int_iff wp_pointwise_discharge)
     apply(auto simp add: mem_act_def full_step_def co_step_def par_step_def thr_pc_def)[1]
  apply simp
    apply(frule store_step_disj)
    apply(erule disjE)
  using assms(3) Int_iff wp_pointwise_discharge
     apply smt
    apply clarsimp
  apply(subgoal_tac "(a, b) \<in> \<langle>full_step Ts Pr (t', Some (store x21 x22 (e (thr_rf (a, b) t'))))\<rangle> (A t (thr_pc (a, b) t))")
  using wp_pointwise_discharge apply fastforce
  apply(subgoal_tac "(a, b) \<in>  A t (thr_pc (a, b) t) \<inter> A t' (thr_pc (a, b) t') \<inter> Reg t' {rf . e rf = e (thr_rf (a, b) t')}")
  using assms(5) apply blast
     apply(fastforce simp add: thr_rf_def lift\<^sub>2_def lift_t_def Reg_def mem_act_def full_step_def co_step_def par_step_def thr_pc_def)[1]

    apply(subgoal_tac "Pr t' (thr_pc (a, b) t') = fencec x31 x32 (thr_pc (aa, ba) t')")
  using assms(4) apply (meson Int_iff wp_pointwise_discharge)
   apply(auto simp add: mem_act_def full_step_def co_step_def par_step_def thr_pc_def)[1]
  
    apply(subgoal_tac "Pr t' (thr_pc (a, b) t') = loada x41 x43 x42 x45 x46 (thr_pc (aa, ba) t')")
  using assms(6) using wp_pointwise_discharge Int_iff apply smt
   apply(auto simp add: mem_act_def full_step_def co_step_def par_step_def thr_pc_def)[1]
    using assms(1)
  apply(subgoal_tac "((a, b), aa, ba) \<in> arb_pc t'")
  apply (meson wp_pointwise_discharge)
    by(fastforce simp add: mem_act_def full_step_def co_step_def par_step_def thr_pc_def wp_ext arb_pc_def)


lemma not_in_Ts_no_step:
  "fst e \<notin> Ts \<Longrightarrow> full_step Ts Pr e = {}"
  apply(cases "snd e")
  by(auto simp add: full_step_def co_step_def par_step_def)

lemma owicki_gries:
  assumes "\<And> t. t \<in> Ts \<Longrightarrow> local_correctness Ts Pr A t"
      and "\<And> t. t \<in> Ts \<Longrightarrow> noninterference Ts Pr A t"
      and "\<And> t s. t \<in> Ts \<Longrightarrow> s \<in> I \<Longrightarrow> s \<in> A t (thr_pc s t)"
    shows "reach (\<Union> e. full_step Ts Pr e) I \<subseteq> inv_of Ts A"
  apply(rule reach_intro1)
  using assms(3) apply(fastforce simp add: inv_of_def)
  apply(unfold inv_of_def) apply simp apply rule
  apply(case_tac "fst e = t")
  using assms(1) apply(fastforce simp add: full_step_def local_correctness_def par_step_def)
  apply rule
  apply(subgoal_tac "thr_pc s' t = thr_pc s t")
   apply(subgoal_tac "s \<in> A t (thr_pc s t)")
  apply(subgoal_tac "fst e \<in> Ts")
   apply(subgoal_tac "s \<in> A (fst e) (thr_pc s (fst e))")
  apply clarsimp using assms(2) noninterference_def
  apply smt
     apply blast using not_in_Ts_no_step apply blast
  apply blast
  by (simp add: full_step_update_other thr_pc_def)

lemma lift\<^sub>1_mono:
  "P \<subseteq> Q \<Longrightarrow> lift\<^sub>1 P \<subseteq> lift\<^sub>1 Q"
  by(auto simp add: lift\<^sub>1_def)

lemma lift\<^sub>2_mono:
  "P \<subseteq> Q \<Longrightarrow> lift\<^sub>2 P \<subseteq> lift\<^sub>2 Q"
  by(auto simp add: lift\<^sub>2_def)

lemma object_act_lifted_wp:
  assumes "\<And> o_act. mem_act a o_act \<Longrightarrow> P \<subseteq> \<langle>G (t, o_act)\<rangle> Q"
  shows "lift\<^sub>1 P \<subseteq> \<langle>full_step Ts Pr (t, Some a)\<rangle> (lift\<^sub>1 Q)"
  using assms by(auto simp add: wp_ext lift\<^sub>1_def co_step_def par_step_def full_step_def)

lemma object_read_lifted_wp:
  assumes "P \<subseteq> \<langle>G (t, R b x v)\<rangle> Q"
  shows "lift\<^sub>1 P \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> (lift\<^sub>1 Q)"
  using assms by(auto simp add: mem_act_def wp_ext lift\<^sub>1_def co_step_def par_step_def full_step_def)

lemma object_read_aux_lifted_wp:
  assumes "P \<subseteq> \<langle>G (t, R b x v)\<rangle> Q"
  shows "lift\<^sub>1 P \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v ra va))\<rangle> (lift\<^sub>1 Q)"
  using assms by(auto simp add: mem_act_def wp_ext lift\<^sub>1_def co_step_def par_step_def full_step_def)

lemma object_write_lifted_wp:
  assumes "P \<subseteq> \<langle>G (t, W b x v)\<rangle> Q"
  shows "lift\<^sub>1 P \<subseteq> \<langle>full_step Ts Pr (t, Some (store b x v))\<rangle> (lift\<^sub>1 Q)"
  using assms by(auto simp add: mem_act_def wp_ext lift\<^sub>1_def co_step_def par_step_def full_step_def)


lemma object_tau_lifted_wp:
  "lift\<^sub>1 P \<subseteq> \<langle>full_step Ts Pr (t, None)\<rangle> (lift\<^sub>1 P)"
  by(auto simp add: wp_ext lift\<^sub>1_def co_step_def par_step_def full_step_def)

lemma client_lifted_wp:
  assumes "P \<subseteq> \<langle>par_step Ts (comm_step Pr) (t, ao)\<rangle> Q"
  shows "lift\<^sub>2 P \<subseteq> \<langle>full_step Ts Pr (t, ao)\<rangle> (lift\<^sub>2 Q)"
  apply(cases ao)
  using assms
  by(auto simp add: wp_ext lift\<^sub>2_def co_step_def full_step_def)

lemma Reg_mono:
  assumes "B \<subseteq> C"
  shows "Reg t B \<subseteq> Reg t C"
  using assms by(auto simp add: Reg_def lift\<^sub>2_def lift_t_def)

lemma Reg_intro:
  assumes "fst(ls t) \<in> B"
  shows "(s, ls) \<in> Reg t B"
  using assms by(simp add: Reg_def lift\<^sub>2_def lift_t_def)

lemma store_wp_Reg_stable:
  shows "(Reg t B) \<subseteq> \<langle>full_step Ts Pr (t', Some (store b x v'))\<rangle> (Reg t B)"
  by(auto simp add: wp_ext Reg_def full_step_def co_step_def par_step_def
  mem_act_def lift\<^sub>2_def lift_t_def )

lemma tau_wp_Reg_stable_self:
  shows "(Reg t B) \<subseteq> \<langle>full_step Ts Pr (t, None)\<rangle> (Reg t B)"
  apply(simp add: wp_ext Reg_def full_step_def co_step_def par_step_def
  mem_act_def lift\<^sub>2_def lift_t_def) by clarsimp

lemma act_wp_Reg_stable_other:
  assumes "t' \<noteq> t"
  shows "(Reg t B) \<subseteq> \<langle>full_step Ts Pr (t', ao)\<rangle> (Reg t B)"
  using assms apply(cases ao)
  by(auto simp add: wp_ext Reg_def full_step_def co_step_def par_step_def
         mem_act_def lift\<^sub>2_def lift_t_def)

lemma load_wp_Reg:
  shows "P \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> (Reg t {reg. reg r = v})"
  by(fastforce simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)


lemma load_aux_wp_Reg:
  shows "P \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v ra va))\<rangle> (Reg t {reg. reg ra = va})"
  by(fastforce simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)

lemma load_aux_wp_Regb:
  assumes "ra \<noteq> r"
  shows "P \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v ra va))\<rangle> (Reg t {reg. reg r = v})"
  using assms by (fastforce simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)

lemma load_wp_Reg_neg1:
  assumes "v' \<noteq> v"
  shows "P \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> (Reg t (-{reg. reg r = v'}))"
  using assms by(fastforce simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)

lemma load_aux_wp_Reg_neg1:
  assumes "v' \<noteq> va"
  shows "P \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v ra va))\<rangle> (Reg t (-{reg. reg ra = v'}))"
  using assms by(fastforce simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)

lemma load_wp_Reg_other:
  assumes "r \<noteq> r'"
  shows "Reg t {reg. reg r' = v'} \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> (Reg t {reg. reg r' = v'})"
  using assms by(auto simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)


lemma load_aux_wp_Reg_other:
  assumes "r \<noteq> r'"
  assumes "ra \<noteq> r'"
  shows "Reg t {reg. reg r' = v'} \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v ra va))\<rangle> (Reg t {reg. reg r' = v'})"
  using assms by(auto simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)

lemma fence_wp_Reg:
  shows "P \<subseteq> \<langle>full_step Ts Pr (t, Some (fence_op r v))\<rangle> (Reg t {reg. reg r = v})"
  by(fastforce simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)

lemma load_wp_Reg2:
  shows "(Reg t B) \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> (Reg t {reg(r := v) | reg. reg \<in> B})"
  by(auto simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)

lemma load_wp_Reg3:
  shows "v\<noteq>u \<Longrightarrow> P \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> (Reg t (-{reg. reg r = u}))"
  by(fastforce simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)


lemma load_wp_Reg4:
  shows "(Reg t (-{reg. reg r = u})) \<subseteq> \<langle>full_step Ts Pr (t, Some (store b x  v))\<rangle> (Reg t (-{reg. reg r = u}))"
  by(fastforce simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)

lemma load_wp_Reg5:
  shows "r\<noteq> r' \<Longrightarrow> (Reg t (-{reg. reg r = u})) \<subseteq> \<langle>full_step Ts Pr (t', Some (load b x r' v))\<rangle> (Reg t (-{reg. reg r = u}))"
  apply(case_tac "t \<noteq> t'")
  apply (simp add: act_wp_Reg_stable_other)
  by(fastforce simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext)

lemma load_wp_Reg6:
  shows "lift\<^sub>1(not_value t x v) \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v'))\<rangle> (Reg t (-{reg. reg r = v}))"
  apply(auto simp add: full_step_def co_step_def mem_act_def par_step_def Reg_def lift\<^sub>2_def lift_t_def
     wp_ext lift\<^sub>1_def not_value_def Tr_def Rds_def)
  by force


lemma load_wp_Reg_neg:
  assumes "P \<subseteq> lift\<^sub>1(\<langle>G (t, R b x v)\<rangle> {})"
  shows "P \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> Q"
  apply(subgoal_tac "P \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> {}")
  using disabled_wp subset_trans apply force
  apply(simp add: full_step_def co_step_def mem_act_def)
  using assms by(auto simp add: wp_ext lift\<^sub>1_def)


lemma load_aux_wp_Reg_neg:
  assumes "P \<subseteq> lift\<^sub>1(\<langle>G (t, R b x v)\<rangle> {})"
  shows "P \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v ra va))\<rangle> Q"
  apply(subgoal_tac "P \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v ra va))\<rangle> {}")
  using disabled_wp subset_trans apply force
  apply(simp add: full_step_def co_step_def mem_act_def)
  using assms by(auto simp add: wp_ext lift\<^sub>1_def)

end

locale RWActions = RWVarsActions + RWVars
begin

lemma load_wp_\<beta>_stable:
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
  shows "lift\<^sub>1 P \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v))\<rangle> (lift\<^sub>1 P)"
  by (simp add: assms object_read_lifted_wp read_preserves_all)


lemma load_aux_wp_\<beta>_stable:
  assumes "P \<subseteq> \<langle>\<beta>\<rangle> P"
  shows "lift\<^sub>1 P \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v ra va))\<rangle> (lift\<^sub>1 P)"
  by (simp add: assms object_read_aux_lifted_wp read_preserves_all)

lemma load_wp_synced_value:
  assumes "v' \<noteq> v"
  shows "lift\<^sub>1(synced t x v) \<subseteq> \<langle>full_step Ts Pr (t, Some (load b x r v'))\<rangle> Q"
  apply(rule subset_trans[where B = "\<langle>full_step Ts Pr (t, Some (load b x r v'))\<rangle> (lift\<^sub>1 {})"])
  apply(rule object_read_lifted_wp)
  using assms not_value_implies_wp 
   apply (smt RWVars.initial_not_value RWVars_axioms R_Rds pres_intro subsetD wp_emp_not_trans)
  apply (unfold lift\<^sub>1_def)
  by (simp add: disabled_wp)

lemma load_aux_wp_synced_value:
  assumes "v' \<noteq> v"
  shows "lift\<^sub>1(synced t x v) \<subseteq> \<langle>full_step Ts Pr (t, Some (load_aux b x r v' ra va))\<rangle> Q"
  apply(rule subset_trans[where B = "\<langle>full_step Ts Pr (t, Some (load_aux b x r v' ra va))\<rangle> (lift\<^sub>1 {})"])
  apply(rule object_read_aux_lifted_wp)
  using assms not_value_implies_wp 
   apply (smt RWVars.initial_not_value RWVars_axioms R_Rds pres_intro subsetD wp_emp_not_trans)
  apply (unfold lift\<^sub>1_def)
  by (simp add: disabled_wp)

lemma reg_neg_empty: "Reg t (- A) \<inter> Reg t (A) \<subseteq> {}"
  apply(auto simp add: Reg_def lift\<^sub>2_def lift_t_def)
  done

lemma reg_neg_P: "(Reg t (- B) \<union> P) \<inter>
       Reg t B \<subseteq> P"
  using reg_neg_empty
  by fastforce

lemma wp_ir_pr_pre_disj1:
   "m \<inter> (p \<union> q) \<subseteq> \<langle>a\<rangle> (r) \<Longrightarrow> m \<inter> p \<subseteq> \<langle>a\<rangle> (r)"
 by(auto simp add: wp_ext)

lemma wp_ir_pr_pre_disj2:
  assumes "m \<inter> p \<subseteq> \<langle>a\<rangle> q"
      and "m \<inter> p' \<subseteq> \<langle>a\<rangle> q"
    shows" m \<inter> (p \<union> p') \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_pre_disj3:
  assumes "p \<inter> m \<subseteq> \<langle>a\<rangle> q"
      and "p' \<inter> m \<subseteq> \<langle>a\<rangle> q"
    shows "(p \<union> p') \<inter> m \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)

lemma wp_ir_pr_conj_elim3:
  assumes "p \<inter> p' \<subseteq> \<langle>a\<rangle> q"
    shows" p \<inter> p' \<inter> p'' \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)


lemma wp_ir_pr_conj_elim4:
  assumes "p \<inter> p'' \<subseteq> \<langle>a\<rangle> q"
    shows" p \<inter> p' \<inter> p'' \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)


lemma wp_ir_pr_conj_elim5:
  assumes "p' \<inter> p'' \<subseteq> \<langle>a\<rangle> q"
    shows" p \<inter> p' \<inter> p'' \<subseteq> \<langle>a\<rangle> q"
  using assms by(auto simp add: wp_ext)


lemma synced_diff_val_diff_threads:  "v \<noteq> v' \<Longrightarrow> t \<noteq> t' \<Longrightarrow> synced t x v \<inter> synced t' x v' \<subseteq> {}"
  apply(simp add: synced_def)
  by (smt Int_absorb2 RWBase.def_value_transfer RWBase_axioms def_value_unique inf.idem inf_assoc inf_bot_right inf_commute)

lemma lift\<^sub>1_inter_empty_mono: "A \<inter> B \<subseteq> {} \<Longrightarrow> lift\<^sub>1 A \<inter> lift\<^sub>1 B \<subseteq> {}"
  by(auto simp add: lift\<^sub>1_def)

lemma synced_diff_threads_vals:  "v \<noteq> v' \<Longrightarrow> t \<noteq> t' \<Longrightarrow> lift\<^sub>1(synced t x v) \<inter> lift\<^sub>1(synced t' x v') \<subseteq> {}"
  using synced_diff_val_diff_threads lift\<^sub>1_inter_empty_mono
  by metis 

end

locale MPSync =  RWActions + MPCase
begin




lemma no_value_stable_diff_var:
  assumes "mem_act ao a"
  and "var a \<noteq> x \<or> wval a \<noteq> Some v"
  shows "lift\<^sub>1 (not_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some ao)\<rangle>
           (lift\<^sub>1 (not_value t x v))"
  using assms
   apply(case_tac "ao")
  apply(simp add: mem_act_def)
  using G_\<beta>_stable
   apply (simp add: load_wp_\<beta>_stable not_value_\<beta>_stable)
  using G_\<beta>_stable
  apply(simp add: mem_act_def)
   apply (simp add: interf_wp2 not_value_\<beta>_stable not_value_stable object_write_lifted_wp)
  apply (metis (mono_tags, lifting) F_wval mem_act_def mem_op.simps(19) not_value_\<beta>_stable object_act_lifted_wp read_preserves_all)
  by (simp add: load_aux_wp_\<beta>_stable not_value_\<beta>_stable)


lemma no_value_stable_diff_var_store:
  assumes  "y \<noteq> x \<or> v\<noteq>u"
  shows "lift\<^sub>1 (not_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some (store b y u))\<rangle>
           (lift\<^sub>1 (not_value t x v))"
  using assms no_value_stable_diff_var[where ao = "store b y u"] mem_act_def
 by (smt W_var W_wval interf_wp2 not_value_\<beta>_stable not_value_stable object_write_lifted_wp option.inject)


lemma constrained_value_stable_diff_var:
  assumes "mem_act ao a"
  and "var a \<noteq> x"
  shows "lift\<^sub>1 (constrained_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some ao)\<rangle>
           (lift\<^sub>1 (constrained_value t x v))"
  using assms
   apply(case_tac "ao")
  apply(simp add: mem_act_def)
  using G_\<beta>_stable
  apply (simp add: constrained_value_\<beta>_stable load_wp_\<beta>_stable)
  using G_\<beta>_stable
   apply(simp add: mem_act_def)
  apply (simp add: constrained_value_\<beta>_stable constrained_value_stable interf_wp object_write_lifted_wp)
  apply (metis (mono_tags, lifting) RWVars.constrained_value_\<beta>_stable RWVars.constrained_value_stable RWVars_axioms interf_wp2 mem_act_def mem_op.simps(19) object_act_lifted_wp)
  by (meson RWVars.constrained_value_\<beta>_stable RWVars_axioms load_aux_wp_\<beta>_stable)


lemma constrained_value_stable_load:
  shows "lift\<^sub>1 (constrained_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some (load b y r u))\<rangle>
           (lift\<^sub>1 (constrained_value t x v))"
  by (simp add: constrained_value_\<beta>_stable load_wp_\<beta>_stable)


lemma constrained_value_stable_store_diff_var:
  assumes "y \<noteq> x"
  shows "lift\<^sub>1 (constrained_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some (store b y u))\<rangle>
           (lift\<^sub>1 (constrained_value t x v))"
  using constrained_value_stable_diff_var  
  by (simp add: constrained_value_stable_diff_var assms mem_act_def)

lemma transfer_wr_diff_var_pres:
  assumes "var w \<noteq> x"
  shows "(not_value t x v) 
      \<subseteq> \<langle>G(t', w)\<rangle> (\<langle>G (t, R True x v)\<rangle> (synced t y u))"
  using assms
  apply(simp add: not_value_def)
  apply(subgoal_tac "(not_value t x v) 
      \<subseteq> \<langle>G(t', w)\<rangle> (not_value t x v) ")
  apply (metis R_Rds TSys.wp_emp_not_trans not_value_def pres_intro wp_pres_compos2)
  by (simp add: interf_wp not_value_\<beta>_stable not_value_stable)


lemma disable_transfer:
  assumes "x \<noteq> y"
  and "mem_act ao a"
  and "var a \<noteq> x"
  shows "lift\<^sub>1 (not_value t x v) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some ao)\<rangle>
           (lift\<^sub>1 (\<langle>G (t, R True x v)\<rangle> (synced t y u)))"
  using assms 
  apply(subgoal_tac "not_value t x v \<subseteq> \<langle>G (t, R True x v)\<rangle> {}")
   apply(case_tac "ao")
  using G_\<beta>_stable
  apply (smt R_Rds R_wval interpolation_left not_value_def object_read_lifted_wp pres_intro read_preserves_all synced_value_\<beta>_stable wp_emp_not_trans)
  apply(simp add: mem_act_def)
  apply(subgoal_tac "not_value t x v
       \<subseteq> \<langle>G(t', a)\<rangle>
           ((\<langle>G (t, R True x v)\<rangle> (synced t y u)))")
    defer
      apply (simp add: transfer_wr_diff_var_pres) 
  apply (smt MPSync.transfer_wr_diff_var_pres MPSync_axioms mem_act_def mem_op.simps(19) object_act_lifted_wp)
  apply (meson disabled_wp lift\<^sub>1_mono load_aux_wp_\<beta>_stable not_value_\<beta>_stable subset_trans wp_pred_mono)
  apply (metis R_Rds not_value_def pres_intro wp_emp_not_trans)
  by (simp add: object_write_lifted_wp)

lemma trasfer_intro: "f \<noteq> x \<Longrightarrow> \<langle>G(t', R True x v')\<rangle>(synced t' f v) \<inter>  synced t f v \<subseteq> \<langle>G(t, W True x v')\<rangle>(\<langle>G(t', R True x v')\<rangle>(synced t' f v))"
  apply(subgoal_tac "W True x v' \<in> releasing") defer
  using releasing_def apply auto[1]
  apply(subgoal_tac " R True x v' \<in> acquiring") defer
  using acquiring_def apply auto[1]
  using \<Gamma>_set_synced[where w = "W True x v'" and r = "R True x v'" and t = t and t'=t' and y = f and v = v]
  by simp

lemma lift\<^sub>1_g: "lift\<^sub>1 P \<inter> lift\<^sub>1 Q \<subseteq> lift\<^sub>1 (P \<inter> Q)"
  by(auto simp add: lift\<^sub>1_def)

lemma lifted_transfer_intro:
   "y \<noteq> x \<Longrightarrow> lift\<^sub>1(\<langle>G (t', R True x v')\<rangle>(synced t' y v)) \<inter> lift\<^sub>1(synced t y v) \<subseteq> \<langle>full_step Ts Pr (t, Some (store True x v'))\<rangle>(lift\<^sub>1(\<langle>G (t', R True x v')\<rangle>(synced t' y v)))"
  using trasfer_intro
  by (smt inf.absorb_iff2 le_infI1 lift\<^sub>1_g object_write_lifted_wp)


lemma cobs_pres: "f \<noteq> x \<Longrightarrow> \<langle>G(t', R True f v')\<rangle>(synced t' f v')  \<subseteq> \<langle>G(t, W b x v')\<rangle>((\<langle>G(t', R True f v')\<rangle>(synced t' f v')))"
  apply(auto simp add: wp_ext synced_def def_value_def \<Gamma>_def not_value_def Var_def Rds_def Tr_def)
  oops

lemma cobs_pres_load:
  shows "lift\<^sub>1(\<langle>G (t, R True x v)\<rangle>(synced t z v')) 
      \<subseteq> \<langle>full_step Ts Pr (t', Some (load b y r u))\<rangle>  
                      (lift\<^sub>1(\<langle>G (t, R True x v)\<rangle> (synced t z v')))"
    apply (simp add: G_\<beta>_stable load_wp_\<beta>_stable synced_value_\<beta>_stable)
  done 

lemma c_obs_write_diff_var_pres:
  assumes "var w \<noteq> x"
  and "var w \<noteq> z"
  shows "\<langle>G (t, R True x v)\<rangle>(synced t z v')
   \<subseteq> \<langle>G(t', w)\<rangle>(\<langle>G (t, R True x v)\<rangle>(synced t z v'))"
  using assms
  by (metis G_\<beta>_stable RWVars.synced_value_stable RWVars_axioms R_var SharedVars.distinct_vars_act_independent SharedVars_axioms interf_wp synced_value_\<beta>_stable wp_lipton)


lemma def_value_enabled_read1: "y \<in> def_value t x v \<Longrightarrow> y \<in> Domain (t \<triangleright> Rds x v)"
  apply(auto simp add: def_value_def not_not_value_domain)
  using Domain_Sup_distrib  read_enabled[where x=x and t=t] 
  using not_not_value_domain
  by (smt ComplD ComplI UNIV_I UN_iff subsetD)

lemma def_value_enabled_read2:
  "def_value t x v \<subseteq> Domain (t \<triangleright> Rds x v)"
  using def_value_enabled_read1
  by (simp add: subsetI)


lemma not_value_def_value_empty:
  shows "not_value t x v \<inter> def_value t x v = {}"
   apply(rule equals0I)
  apply(subgoal_tac "y \<in> not_value t x v")
   defer using def_value_implies_not_value 
   apply simp
  apply(subgoal_tac "y \<in> def_value t x v")
   defer apply simp
  apply(subgoal_tac "y \<in> Domain (t \<triangleright> Rds x v)")
   defer
   apply (simp add: def_value_enabled_read1)
  by (simp add: ComplD not_not_value_domain)

lemma not_value_synced_empty: 
    "not_value t x v \<inter> synced t x v \<subseteq> {}"
  using not_value_def_value_empty synced_def by fastforce

end


end
