theory TSO
  imports UnifiedAxioms HOL.Rat
begin

(*********** Basic types ***********)

type_synonym T = nat (* Thread ID *)
type_synonym L = nat (* Location *)
 
definition "null = (0 :: nat)"

type_synonym V = nat

datatype action =
    Read L V
  | Write L V
  | Fence

fun avar :: "action \<Rightarrow> L"
  where
    "avar (Read x v) = x"
  | "avar (Write x v) = x"

fun wval :: "action \<Rightarrow> V option"
  where
    "wval (Read x v) = None"
  | "wval (Write x v) = Some v"
  | "wval Fence = None"

fun rval :: "action \<Rightarrow> V option"
  where
    "rval (Read x v) = Some v"
  | "rval (Write x v) = None"
  | "rval Fence = None"

abbreviation "just_write a \<equiv> wval a \<noteq> None \<and> rval a = None"


(*********** Pure TSO ***********)

type_synonym tso_state = "(L \<Rightarrow> V) \<times> (T \<Rightarrow> (L \<times> V) list)"

definition "lbuffer w = {((s, (buf @ [w])), (s(fst w := snd w), buf) ) | buf s. True}"

fun buffer :: "T \<times> (L \<times> V) \<Rightarrow> tso_state rel"
  where
  "buffer (t, w) =
    {((s, bufs), (s', bufs(t := buf'))) | s bufs s' buf'.
            ((s, bufs(t)), (s', buf')) \<in> lbuffer w}"


primrec rd_buf :: "(L \<times> V) list \<Rightarrow> L \<Rightarrow> V option"
  where
    "rd_buf [] x = None"
  | "rd_buf (h # t) x = (if fst h = x then Some (snd h) else rd_buf t x)"

definition "read st buf x \<equiv> case (rd_buf buf x) of Some v \<Rightarrow> v | None \<Rightarrow> st x"

(* Strictly speaking, the following is redundant but it's good to know. *)
lemma read_latest:
  "rd_buf ((x, v) # buf) x = Some v"
  by simp

lemma rd_buf_app_rd_write [simp]:
  "rd_buf b x = None \<Longrightarrow> x = fst w \<Longrightarrow> rd_buf (b  @ [w]) x = Some (snd w)"
  apply(induct b)
  by auto

lemma rd_buf_append_no_rd [simp]:
  "rd_buf b x = None \<Longrightarrow> x \<noteq> fst w \<Longrightarrow> rd_buf (b  @ [w]) x = None"
  apply(induct b)
  by auto

lemma rd_buf_append_earlier_write [simp]:
  "rd_buf b x = Some v \<Longrightarrow> rd_buf (b  @ [w]) x = Some v"
  apply(induct b)
  by auto

lemma rd_buf_own_write_match:
  "x = fst w \<Longrightarrow> rd_buf (w#b) x = Some (snd w)"
  by simp

lemma rd_buf_own_write_no_match:
  "x \<noteq> fst w \<Longrightarrow> rd_buf (w#b) x = rd_buf b x"
  by simp

lemma flush_read_invisible:
  assumes "((s, b), (s', b')) \<in> lbuffer w"
  shows "read s b x = read s' b' x"
  using assms
  apply(simp add: read_def lbuffer_def)
  apply(cases "rd_buf b' x")
  by auto

definition "internal \<equiv> \<Union> t w. buffer (t, w)"

fun action_step :: "T \<times> action \<Rightarrow> tso_state rel"
  where
    "action_step (t, Read x v) = {((s, bs), (s, bs)) | s bs . v = read s (bs t) x}"
  | "action_step (t, Write x v) = {((s, bs), (s, bs(t := (x, v)#(bs t)))) | s bs . True}"
  | "action_step (t, Fence) = {((s, bs), (s, bs)) | s bs .bs t = []}"

definition "tso_I = {(s, \<lambda> t. []) | s. True}"


lemma distinct_threads_vars_buffer_commutes:
  assumes "fst w \<noteq> fst w'"
      and "t \<noteq> t'"
    shows "buffer (t, w) O buffer (t', w') = buffer (t', w') O buffer (t, w)"
  using assms by(auto simp add: lbuffer_def fun_upd_twist relcomp.simps)

lemma write_action_buffer_right_move_other_threads:
  assumes "t \<noteq> t'"
  shows "action_step (t, Write x v) O buffer (t', w) \<subseteq> buffer (t', w) O action_step (t, Write x v)"
  using assms by(auto simp add: lbuffer_def fun_upd_twist relcomp.simps)



(*********** Prophetic TSO ***********)

record proph_tso_state =
  my_tso :: tso_state
  abufs :: "T \<Rightarrow> rat list"


fun proph_buffer :: "rat \<Rightarrow> T \<times> (L \<times> V) \<Rightarrow> proph_tso_state rel"
  where
  "proph_buffer q (t, w) =
    {(pr, \<lparr> my_tso = tso', abufs = (abufs pr)(t := ab)\<rparr>) |
          pr tso' ab.
      (my_tso pr, tso') \<in> buffer (t, w) \<and>
      (\<forall> t' q'. t' \<noteq> t \<longrightarrow> q' \<in> set (abufs pr t') \<longrightarrow> q < q') \<and>
      abufs pr t = ab @ [q]}"

(* Check this abufs freshness condition *)
definition
  "proph_write_ts q t x v =
    {(pr, \<lparr> my_tso = tso', abufs = (abufs pr)(t := q#(abufs pr t))\<rparr>) |
          pr tso'.
      (my_tso pr, tso') \<in> action_step (t, Write x v) \<and>
      (\<forall> q'. q' \<in> set (abufs pr t) \<longrightarrow> q' < q) \<and>
      (\<forall> t'. q \<notin> set (abufs pr t'))}"


lemma proph_write_ts_valid_for_self:
  "(pr, pr') \<in> proph_write_ts q t x v \<Longrightarrow> abufs pr t = [] \<or> (abufs pr t) ! 0 < q"
  apply(simp add: proph_write_ts_def)
  using nth_mem by auto

lemma proph_write_ts_valid_for_all:
  "(pr, pr') \<in> proph_write_ts q t x v \<Longrightarrow> q \<notin> set (abufs pr t')"
  by(auto simp add: proph_write_ts_def)

definition "on_tso e =
     {(pr, \<lparr>my_tso = tso', abufs = abufs pr\<rparr>) | pr tso'. (my_tso pr, tso') \<in> action_step e}"

fun proph_action :: "T \<times> action \<Rightarrow> proph_tso_state rel"
  where
    "proph_action (t, Read x v) = on_tso (t, Read x v)"
  | "proph_action (t, Write x v) = (\<Union> q. proph_write_ts q t x v)"
  | "proph_action (t, Fence) = on_tso (t, Fence)"

definition "proph_internal = (\<Union> q t w . proph_buffer q (t, w))"


definition "proph_step e = (\<Union> q t w . proph_buffer q (t, w))\<^sup>* O proph_action e"


definition "proph_I = {\<lparr> my_tso = tso, abufs = (\<lambda> t. [])\<rparr> | tso. tso \<in> tso_I}"

definition valid_ts_buff :: "rat list \<Rightarrow> bool"
  where
  "valid_ts_buff ab \<equiv>
      (\<forall> m n. m < n \<and> n < length ab \<longrightarrow> nth ab n < nth ab m)"

lemma valid_ts_buff_injective:
  assumes "valid_ts_buff ab"
    and "m < length ab"
    and "n < length ab"
    and "nth ab m = nth ab n"
  shows "m = n"
  apply(case_tac "ab = []")
  using assms apply(auto simp add: valid_ts_buff_def)[1]
  apply(case_tac "m < n")
  using assms apply(fastforce simp add: valid_ts_buff_def)
  apply(case_tac "n < m")
  using assms apply(fastforce simp add: valid_ts_buff_def)
  by(simp add: valid_ts_buff_def)

lemma valid_ts_buff_emp:
  "valid_ts_buff []"
  by(simp add: valid_ts_buff_def)

lemma larger_hd_largest_all:
  "valid_ts_buff ab \<Longrightarrow> ab \<noteq> [] \<Longrightarrow> hd ab < q \<Longrightarrow> n < length ab \<Longrightarrow> ab ! n < q"
  apply(auto simp add: valid_ts_buff_def)
  by (metis dual_order.strict_trans hd_conv_nth less_Suc_eq less_Suc_eq_0_disj)

lemma valid_ts_buff_head_notin_tail:
  assumes "valid_ts_buff ab"
     and "ab \<noteq> []"
   shows " hd ab \<notin> set (tl ab)"
  using assms apply(auto simp add: valid_ts_buff_def)
  by (smt Nitpick.size_list_simp(2) Suc_mono assms(1) hd_conv_nth in_set_conv_nth length_pos_if_in_set less_SucE not_less_zero nth_tl valid_ts_buff_injective zero_induct)


lemma valid_ts_buff_push:
  assumes "ab = [] \<or> ab ! 0 < q"
  shows "valid_ts_buff ab \<longleftrightarrow> valid_ts_buff (q#ab)"
  apply(case_tac "ab = []")
   apply(fastforce simp add: valid_ts_buff_def)
  using assms apply (simp add: hd_conv_nth)
  apply rule
   apply(simp add: valid_ts_buff_def)
  apply clarsimp
   apply(case_tac "m = 0") apply simp 
  apply (smt Suc_lessI cancel_comm_monoid_add_class.diff_cancel diff_Suc_less dual_order.strict_trans less_Suc_eq zero_less_diff) 
  using less_Suc_eq_0_disj apply fastforce
  by (smt Suc_mono dual_order.strict_trans length_Cons list.sel(3) nth_tl valid_ts_buff_def)

lemma valid_ts_buff_flush:
  assumes "\<forall> q'. q' \<in> set ab \<longrightarrow> q < q'"
  shows "valid_ts_buff ab \<longleftrightarrow> valid_ts_buff (ab@[q])"
  apply(case_tac "ab = []")
   apply(simp add: valid_ts_buff_def)
  using assms apply(simp add: valid_ts_buff_def nth_append)
  by blast

definition
  "valid_buffers bb ab \<equiv>
      length bb = length ab \<and> valid_ts_buff ab"

lemma valid_buffers_both_emp:
  assumes "valid_buffers bb ab"
  shows "bb = [] \<longleftrightarrow> ab = []"
  using assms by(auto simp add: valid_buffers_def)

lemma valid_buffers_emp:
  shows "valid_buffers [] []"
  by(simp add: valid_buffers_def valid_ts_buff_def)

lemma valid_buffers_push:
  assumes "ab = [] \<or> ab ! 0 < q"
  shows "valid_buffers bb ab \<longleftrightarrow> valid_buffers (w#bb) (q#ab)"
  using assms apply(simp add: valid_buffers_def)
  using valid_ts_buff_push
  by blast

lemma valid_buffers_flush:
  assumes "\<forall> q'. q' \<in> set ab \<longrightarrow> q < q'"
    shows "valid_buffers bb ab \<longleftrightarrow> valid_buffers (bb@[w]) (ab@[q])"
  using assms apply(simp add: valid_buffers_def)
  using valid_ts_buff_flush
  by blast

definition valid_state :: "proph_tso_state \<Rightarrow> bool"
  where
  "valid_state pr \<equiv>
      (\<forall> t. valid_buffers (snd (my_tso pr) t) (abufs pr t)) \<and>
      (\<forall> t t'. t \<noteq> t' \<longrightarrow> set (abufs pr t) \<inter> set (abufs pr t') = {})"

definition "valid_fin_state ps \<equiv> valid_state ps \<and> finite {t . abufs ps t \<noteq> []}"

definition
  "writes_vals bb ab \<equiv> set(zip bb ab)"

lemma writes_vals_emp:
  shows "writes_vals [] [] = {}"
  by(auto simp add: writes_vals_def)

lemma writes_vals_push:
  assumes "valid_buffers bb ab"
  shows "writes_vals (w#bb) (q#ab) = writes_vals bb ab \<union> {(w, q)}"
  by(auto simp add: writes_vals_def)

lemma writes_vals_flush:
  assumes "valid_buffers (bb@[w]) (ab@[q])"
  shows "writes_vals (bb@[w]) (ab@[q]) = writes_vals bb ab \<union> {(w, q)}"

  apply(subgoal_tac "zip (bb@[w]) (ab@[q]) = (zip bb ab)@[(w, q)]")
   apply(fastforce simp add: writes_vals_def less_Suc_eq_0_disj)
  using assms valid_buffers_def zip_append by fastforce

definition
  "writes bb ab \<equiv> {(x, q) | x v q. ((x, v), q) \<in> writes_vals bb ab}"

lemma write_value_exists:
  assumes "(x, q) \<in> writes bb ab"
  shows "\<exists> v. ((x, v), q) \<in> writes_vals bb ab"
  using assms by(auto simp add: writes_def)

lemma unique_write_for_write:
  assumes "valid_buffers bb ab"
   and "(x, q) \<in> writes bb ab"
 shows "\<exists>! v. ((x, v), q) \<in> writes_vals bb ab"
  apply (rule ex_ex1I[OF write_value_exists[OF assms(2)]])
  using assms(1)
  apply( simp add: valid_buffers_def writes_vals_def)
  using valid_ts_buff_injective
  by (metis in_set_zip prod.sel(1) prod.sel(2))

definition
      "proph_buf_pre t tso pr' q \<equiv>
       \<lparr> my_tso = tso,
         abufs = (abufs pr')(t := (abufs pr' t) @ [q])\<rparr>"

definition
      "proph_write_pre t tso pr' \<equiv>
       \<lparr> my_tso = tso,
         abufs = (abufs pr')(t := tl (abufs pr' t))\<rparr>"

definition
      "proph_act_pre tso pr' = 
       \<lparr> my_tso = tso,
         abufs = abufs pr'\<rparr>"

definition bsim :: "tso_state \<Rightarrow> proph_tso_state \<Rightarrow> bool"
  where
  "bsim tso pr \<equiv> tso = my_tso pr \<and> valid_state pr"

lemma proph_buffer_on_tso:
  assumes "(pr, pr') \<in> proph_buffer q' (t, w)"
  shows "(my_tso pr, my_tso pr') \<in> buffer (t, w)"
  using assms by force

lemma buffer_bsim_valid_transition:
  assumes "bsim tso' pr'"
    and "(tso, tso') \<in> buffer (t, w)"
     and "\<forall> t' q. t' \<noteq> t \<longrightarrow> q \<in> set (abufs pr' t') \<longrightarrow> q' < q"
  shows "(proph_buf_pre t tso pr' q', pr') \<in> proph_buffer q' (t, w)"
  using assms apply (simp add: bsim_def proph_buf_pre_def)
  apply clarsimp
  by (metis old.unit.exhaust surjective)

lemma valid_ts_buff_q_least:
  assumes "valid_ts_buff (ab @ [q])"
    and "q' \<in> set ab"
  shows "q < q'"
  using assms apply(auto simp add: valid_ts_buff_def)
  by (metis butlast_snoc in_set_conv_nth lessI nth_append_length nth_butlast)


lemma disjoint_pointwise:
  shows "X \<inter> Y = {} \<longleftrightarrow> (\<forall> x. x \<in> X \<longrightarrow> x \<notin> Y )"
  by auto

lemma proph_buffer_least_q_other:
  assumes "valid_state pr"
     and "(pr, pr') \<in> proph_buffer q (t, w)"
     and "q' \<in> set (abufs pr t')"
     and "t' \<noteq> t"
   shows "q < q'"
  using assms by(auto simp add: valid_state_def valid_buffers_def valid_ts_buff_def)


lemma proph_buffer_post_valid:
  assumes "valid_state pr"
     and "(pr, pr') \<in> proph_buffer q' (t, w)"
   shows "valid_state pr'"
  apply(subgoal_tac "length (snd (my_tso pr) t) = length (abufs pr t)")
  defer using assms(1) valid_buffers_def valid_state_def apply blast
  apply(simp add: valid_state_def)
  apply(rule conjI)
   apply(rule)
  apply(case_tac "ta = t")
   apply(simp add: valid_buffers_def)
    apply(rule conjI)
  using assms(2) apply (simp add: lbuffer_def) apply fastforce
  using assms(2) apply simp 
  using valid_ts_buff_flush valid_ts_buff_q_least
  apply (metis assms(1) fun_upd_same select_convs(2) valid_buffers_def valid_state_def)
  apply(subgoal_tac "length (snd (my_tso pr) ta) = length (abufs pr ta)")
  defer using assms(1) valid_buffers_def valid_state_def apply blast
    defer 
   apply(simp add: valid_buffers_def)
  apply(rule conjI) using assms(2) apply simp
    apply fastforce
  using assms(2) apply simp
   apply (metis assms(1) fun_upd_other select_convs(2) valid_buffers_def valid_state_def)
  apply(unfold disjoint_pointwise)
  apply rule apply rule apply rule apply rule apply rule
  apply(case_tac "t' = t")
  using assms(1, 2) apply (simp add: valid_state_def) apply fastforce
  apply(case_tac "ta = t")
  using assms(1, 2) apply (simp add: valid_state_def) apply fastforce
  using assms(1, 2) apply (simp add: valid_state_def) by fastforce

lemma proph_internal_post_valid:
  assumes "valid_state pr"
     and "(pr, pr') \<in> proph_internal"
   shows "valid_state pr'"
  using assms apply(unfold proph_internal_def)
  using proph_buffer_post_valid
  by blast


lemma abufs_proph_buffer:
  "(pr, pr') \<in> proph_buffer q (t, w) \<Longrightarrow> abufs pr t = abufs pr' t @ [q]"
  by auto

lemma proph_buffer_pre_valid:
  assumes "valid_state pr'"
     and "(pr, pr') \<in> proph_buffer q' (t, w)"
     and "\<And> t' q. q \<in> set (abufs pr' t') \<Longrightarrow> q' < q"
   shows "valid_state pr"
     apply(subgoal_tac "abufs pr t = abufs pr' t @[q']")
  defer using assms(2) apply fastforce 
  apply(subgoal_tac "snd (my_tso pr) t = (snd (my_tso pr') t) @ [w]")
  defer using assms(2) apply (simp add: lbuffer_def)
  apply (metis fun_upd_same select_convs(1) snd_conv)
  using assms(2)
  apply(simp add: valid_state_def)
  apply rule
   apply rule
   apply(case_tac "ta = t")
  using valid_buffers_flush
  using assms(3) apply (metis assms(1) valid_state_def)
  apply (metis assms(1) fun_upd_same fun_upd_triv fun_upd_twist select_convs(1) select_convs(2) snd_conv valid_state_def)
  apply clarsimp
  apply(subgoal_tac "set (abufs pr t) = set (abufs pr' t) \<union> {q'}")
  defer using abufs_proph_buffer[OF assms(2)] apply fastforce defer
  apply(case_tac "t' = t")
  using assms(1) apply(simp add: valid_state_def) 
  apply (metis assms(3) fun_upd_same fun_upd_triv fun_upd_twist less_irrefl select_convs(2))
  apply(case_tac "ta = t")
  using assms(1) apply(simp add: valid_state_def)
  apply (metis assms(3) fun_upd_other less_irrefl select_convs(2))
  by (smt assms(1) fun_upd_same fun_upd_triv fun_upd_twist select_convs(2) valid_state_def)


lemma valid_fin_state_proph_buffer_pres:
  assumes "valid_fin_state ps"
  assumes "(ps, ps') \<in> proph_buffer q (t, w)"
  shows "valid_fin_state ps'"
  using assms(1) apply(unfold valid_fin_state_def)
  apply(rule conjI)
  using assms(2) proph_buffer_post_valid apply blast
  apply(subgoal_tac "{t. abufs ps' t \<noteq> []} \<subseteq> {t. abufs ps t \<noteq> []}")
  using finite_subset apply blast
  using assms(2) apply auto
  by (metis Nil_is_append_conv)


lemma proph_internal_post_fin_valid:
  assumes "valid_fin_state pr"
     and "(pr, pr') \<in> proph_internal"
   shows "valid_fin_state pr'"
  using assms(1) apply(unfold valid_fin_state_def)
  apply(rule conjI)
  using assms(2) proph_internal_post_valid apply blast
  by (metis (no_types, lifting) UN_iff assms(2) proph_internal_def valid_fin_state_def valid_fin_state_proph_buffer_pres)

lemma buffer_bsim:
  assumes "bsim tso' pr'"
    and "(tso, tso') \<in> buffer (t, w)"
     and "\<And> t' q. q \<in> set (abufs pr' t') \<Longrightarrow> q' < q"
  shows "bsim tso (proph_buf_pre t tso pr' q')"
  apply(unfold bsim_def)
  apply rule
   apply (fastforce simp add: proph_buf_pre_def)
  using buffer_bsim_valid_transition[OF assms(1) assms(2)] proph_buffer_pre_valid
  using assms(1) assms(3) bsim_def by blast

lemma proph_write_on_tso:
  assumes "(pr, pr') \<in> proph_write_ts q' t x v"
  shows "(my_tso pr, my_tso pr') \<in> action_step (t, Write x v)"
  using assms by (auto simp add: proph_write_ts_def)


lemma write_bsim_valid_transition:
  assumes "bsim tso' pr'"
    and "(tso, tso') \<in> action_step (t, Write x v)"
  shows "(proph_write_pre t tso pr', pr') \<in> proph_action (t, Write x v)"
  apply(subgoal_tac "(proph_write_pre t tso pr', pr') \<in> proph_write_ts (hd (abufs pr' t)) t x v")
   apply fastforce
  apply(subgoal_tac "abufs pr' t \<noteq> []")
   defer using assms apply(simp add: bsim_def valid_state_def valid_buffers_both_emp)
  apply (metis fun_upd_same list.discI snd_conv valid_buffers_both_emp)
  apply(simp add: proph_write_ts_def proph_write_pre_def)
  apply(rule exI[where x="fst (my_tso pr')"])
  apply(rule exI[where x="snd (my_tso pr')"])
  apply safe
      apply fastforce
  using assms(1) assms(2) bsim_def apply fastforce
  apply (smt Nitpick.size_list_simp(2) Suc_mono assms(1) bsim_def in_set_conv_nth larger_hd_largest_all less_irrefl linorder_neqE_linordered_idom nth_tl valid_buffers_def valid_state_def valid_ts_buff_head_notin_tail)
  using assms valid_ts_buff_head_notin_tail apply(simp add: bsim_def)
  using valid_buffers_def valid_state_def apply blast
  using assms apply(simp add: bsim_def valid_state_def)
  by fastforce

lemma proph_write_ts_post_valid:
  assumes "valid_state pr"
     and "(pr, pr') \<in> proph_write_ts q' t x v"
   shows "valid_state pr'"
  apply(subgoal_tac "snd (my_tso pr') t = (x,v)#snd (my_tso pr) t")
  defer using assms(2) proph_write_on_tso apply fastforce
  apply(subgoal_tac "abufs pr' t = q'#(abufs pr t)")
  defer using assms(2) apply(fastforce simp add: proph_write_ts_def)
  apply(simp add: valid_state_def)
  apply rule apply rule
  using proph_write_on_tso[OF assms(2)]
  apply(case_tac "ta = t")
   using valid_buffers_push[OF proph_write_ts_valid_for_self[OF assms(2)]]
   using assms(1) valid_state_def apply fastforce
   using assms(1, 2) apply(fastforce simp add: proph_write_ts_def valid_state_def)
   apply rule apply rule
  apply(case_tac "ta = t")
   using assms(1, 2) apply(fastforce simp add: proph_write_ts_def valid_state_def)
  apply(case_tac "t' = t")
   using assms(1, 2) apply(fastforce simp add: proph_write_ts_def valid_state_def)
  using assms(1, 2) by(fastforce simp add: proph_write_ts_def valid_state_def)

lemma proph_write_ts_pre_valid:
  assumes "valid_state pr'"
     and "(pr, pr') \<in> proph_write_ts q' t x v"
   shows "valid_state pr"
  apply(subgoal_tac "snd (my_tso pr') t = (x,v)#snd (my_tso pr) t")
  defer using assms(2) proph_write_on_tso apply fastforce
  apply(subgoal_tac "abufs pr' t = q'#(abufs pr t)")
  defer using assms(2) apply(fastforce simp add: proph_write_ts_def)
  apply(simp add: valid_state_def)
  apply rule
   apply rule
   apply(case_tac "ta = t")
  apply simp
  apply (metis assms(1) assms(2) proph_write_ts_valid_for_self valid_buffers_push valid_state_def)
  using assms proph_write_on_tso[OF assms(2)] apply(simp add: proph_write_ts_def)
  apply (metis fun_upd_same fun_upd_triv fun_upd_twist select_convs(2) snd_conv valid_state_def)
  apply rule apply rule
  apply(case_tac "ta = t")
   using assms(1, 2) apply(fastforce simp add: proph_write_ts_def valid_state_def)
  apply(case_tac "ta = t")
   using assms(1, 2) apply(fastforce simp add: proph_write_ts_def valid_state_def)
  using assms(1, 2) by(fastforce simp add: proph_write_ts_def valid_state_def)

lemma write_bsim:
  assumes "bsim tso' pr'"
    and "(tso, tso') \<in> action_step (t, Write x v)"
   shows "bsim tso (proph_write_pre t tso pr')"
  apply(simp add: bsim_def)
  apply rule
   apply(fastforce simp add: proph_write_pre_def)
  using proph_write_ts_pre_valid write_bsim_valid_transition[OF assms(1) assms(2)]
  using assms(1) apply(simp add: bsim_def)
  by blast

lemma bsim_intro:
  assumes "bsim c a"
      and "c \<in> tso_I"
    shows "a = \<lparr>my_tso = c, abufs = \<lambda>t. []\<rparr>"
  using assms apply(simp add: bsim_def tso_I_def valid_state_def)
  apply safe
  apply(subgoal_tac "abufs a = (\<lambda>t. [])")
   apply fastforce
  apply(rule fun_ext)
  using valid_buffers_both_emp by fastforce

lemma bsim_prestate_on_tso:
  assumes "bsim c' a'"
      and "(c, c') \<in> action_step e"
      and "\<not> just_write (snd e)"
    shows  "(a'\<lparr>my_tso := c\<rparr>, a') \<in> on_tso e"
  apply(cases "snd e")
  using assms apply(auto simp add: bsim_def on_tso_def)
  by (metis (full_types) old.unit.exhaust prod.collapse surjective)+

lemma bsim_on_tso:
  assumes "bsim c' a'"
      and "(c, c') \<in> action_step e"
      and "\<not> just_write (snd e)"
    shows  "bsim c (a'\<lparr>my_tso := c\<rparr>)"
  apply(cases "snd e")
  using assms apply(auto simp add: bsim_def on_tso_def valid_state_def)
  apply (smt CollectD action_step.simps(1) pair_in_Id_conv prod.collapse)
   by (smt CollectD action_step.simps(3) prod.collapse prod.sel(1) snd_conv)

definition
  "the_q t tso pr \<equiv>
     if (\<Union> t'. set (abufs pr t')) = {} then
         (1 :: rat)
     else
         Min (\<Union> t'. set (abufs pr t')) - 1"

lemma Min_Sup:
  assumes "finite (\<Union> a. X a)"
      and "X a \<noteq> {}"
    shows "Min (\<Union> a. X a) \<le> Min (X a)"
  apply(subgoal_tac "(\<Union> a. X a) \<noteq> {}")
   defer using assms(2) apply blast
  apply(subgoal_tac "finite (X a)")
   defer apply(rule rev_finite_subset[OF assms(1)]) apply blast
  apply(subgoal_tac "Min (X a) \<in> X a")
   defer  using Min_in assms(2) apply blast
  by (metis Min.subset_imp UNIV_I UN_upper assms(1) assms(2))

lemma Min_Sup2:
  assumes "finite (\<Union> a. X a)"
      and "q \<in> X a"
    shows "Min (\<Union> a. X a) \<le> Min (X a)"
  using Min_Sup
  using assms(1) assms(2) empty_iff by fastforce

lemma the_q_global_min:
  assumes "q \<in> set (abufs pr t')"
      and "finite (\<Union> t'. set (abufs pr t'))"
    shows "(the_q t tso pr) < q"
  apply(subgoal_tac "Min (\<Union> t'. set (abufs pr t')) \<le> Min (set (abufs pr t'))")
   defer using Min_Sup2[OF assms(2) assms(1)] apply blast
  apply(subgoal_tac "finite (set (abufs pr t'))")
  defer apply blast
  apply(subgoal_tac "Min (\<Union> t'. set (abufs pr t')) \<le> q")
  defer apply (meson Min_le assms(1) dual_order.trans)
  apply(simp add: the_q_def) apply safe
  using assms by fastforce

lemma finitely_nonempty_finite_Un:
  assumes "\<And> x. finite (f x)"
    and "finite {x . f x \<noteq> {}}"
  shows "finite (\<Union> (range f))"
  apply(subgoal_tac "(\<Union> (range f)) = (\<Union> x \<in> {x . f x \<noteq> {}}. f x)")
  using finite_Union apply (fastforce simp add: assms(1) assms(2))
  by auto

lemma finite_Un2:
  assumes "\<And> x. x \<in> X \<Longrightarrow> finite (F x)"
    and "finite {x . F x \<noteq> {}}"
  shows "finite (\<Union> x \<in> X. F x)"
  oops


lemma internal_bsim_q:
  assumes "bsim c' a'"
   and "(c, c') \<in> buffer (t, w)"
   and "finite {t . abufs a' t \<noteq> []}"
 shows "(proph_buf_pre t c a' (the_q t c a'), a') \<in> proph_buffer (the_q t c a') (t, w) \<and>
        bsim c (proph_buf_pre t c a' (the_q t c a'))"
  apply(rule conjI)
   apply(rule buffer_bsim_valid_transition[OF assms(1) assms(2)])
  apply clarsimp
  apply(rule the_q_global_min) apply blast
  apply (smt Collect_cong List.finite_set assms(3) finitely_nonempty_finite_Un set_empty2)
  apply(rule buffer_bsim[OF assms(1) assms(2)])
  apply(erule the_q_global_min)
  apply(rule finitely_nonempty_finite_Un)
  apply fastforce using assms(3) by auto


lemma started_threads_finite:
  "reach (trel action_step internal) tso_I \<subseteq> {s . finite {t . snd s t \<noteq> []}}"
  apply(rule reach_intro)
    apply(simp add: tso_I_def) apply fastforce
  apply(case_tac "snd e") apply simp_all
      apply fastforce
     apply(subgoal_tac "{t. snd s' t \<noteq> []} \<subseteq> {t. snd s t \<noteq> []} \<union> {fst e}")
      apply (fastforce simp add: finite_subset)
     apply fastforce
     apply(subgoal_tac "{t. snd s' t \<noteq> []} = {t. snd s t \<noteq> []}")
     apply fastforce
  apply force
     apply(subgoal_tac "{t. snd s' t \<noteq> []} \<subseteq> {t. snd s t \<noteq> []}")
      apply (fastforce simp add: finite_subset)
  apply(simp add: internal_def lbuffer_def)
  by (metis (mono_tags, lifting) Collect_mono_iff Nil_is_append_conv fun_upd_other list.discI snd_conv)

lemma bsim_started_set_eq:
  assumes "bsim c a"
  shows "{t . snd c t \<noteq> []} = {t . abufs a t \<noteq> []}"
  using assms apply(simp add: bsim_def valid_state_def )
  using valid_buffers_both_emp by fastforce

lemma internal_bsim:
  assumes "bsim c' a'"
   and "(c, c') \<in> buffer (t, w)"
   and "c \<in> reach (trel action_step internal) tso_I"
 shows "\<exists>a. (a, a') \<in> proph_internal \<and> bsim c a"
  apply(subgoal_tac "finite {t. abufs a' t \<noteq> []}")
   apply(rule exI[where x="proph_buf_pre t c a' (the_q t c a')"])
  apply(unfold proph_internal_def)
  apply(rule conjI)
  using internal_bsim_q[OF assms(1) assms(2)] apply simp
    apply (metis prod.collapse)
  using internal_bsim_q[OF assms(1) assms(2)] apply fastforce
  apply(subgoal_tac "finite {t. snd c' t \<noteq> []}")
  using bsim_started_set_eq[OF assms(1)] apply fastforce
  apply(subgoal_tac "c' \<in> reach (trel action_step internal) tso_I")
  using started_threads_finite apply blast
  apply(rule reach_internal[OF assms(3)])
  using assms(2) apply(unfold internal_def) by blast


lemma proph_tso_backsim:
  "backsim bsim action_step internal tso_I proph_action proph_internal proph_I"
  apply(rule backsim_intro)
  apply(simp add: tso_I_def proph_I_def)
  using bsim_intro tso_I_def apply fastforce
  apply(case_tac "snd e")
  using bsim_on_tso bsim_prestate_on_tso
  apply (metis prod.collapse proph_action.simps(1) wval.simps(1))
  using write_bsim write_bsim_valid_transition apply fastforce
  apply (metis bsim_on_tso bsim_prestate_on_tso prod.collapse proph_action.simps(3) rval.simps(3) wval.simps(3))
  using internal_bsim apply(unfold internal_def) by blast

end

